# Star 2D

## Overview

Star-2D is a C library that manages a virtual 2D environment composed of line
segments. The resulting virtual scene can then be accessed through
*ray-tracing*, *uniform sampling* or *closest point query*, providing an
efficient way to deal with geometric data of arbitrary 2D contents.  To ensure
the efficiency of these operators, Star-3D internally relies on [Intel(R)
Rendering Framework](https://software.intel.com/en-us/rendering-framework):
[Embree](https://embree.github.io) that provides highly optimized acceleration
structures as well as traversal kernels for a wide range of data workloads.

To query the scene data one has to create a *view* of the scene. On its
creation, the view internally builds data structures required by the
aforementioned access operators. These data structures are built from the scene
geometry as defined at the moment of the view creation; a view is thus
insensitive to scene updates following its creation. This means that several
views can be used to register different states of the same scene, giving to the
caller a great flexibility to manage the scene data.

## Install

### Pre-requisites

Star-2D is compatible GNU/Linux as well as Microsoft Windows 7 and later, both
in 64-bits. It was successfully built with the [GNU Compiler
Collection](https://gcc.gnu.org) (versions 4.7 and later) as well as with
Microsoft Visual Studio 2015. It relies on [CMake](http://www.cmake.org) and the
[RCMake](https://gitlab.com/vaplv/rcmake/) package to build. It also depends on
the [RSys](https://gitlab.com/vaplv/rsys/) and
[Embree](https://embree.github.io/) libraries.

### How to build

First ensure that CMake and a C/C++ compiler is installed on your system. Then
install the [RCMake](https://gitlab.com/vaplv/rcmake.git) package as well as
the [RSys](https://gitlab.com/vaplv/rsys.git) and
[Embree](https://embree.github.io) libraries. Finally generate the project from
the `cmake/CMakeLists.txt` file by appending to the `CMAKE_PREFIX_PATH`
variable the `<RCMAKE_DIR>`, `<RSYS_DIR>` and `<EMBREE_DIR>`
directories, where `<RCMAKE_DIR>`, `<RSYS_DIR>` and `<EMBREE_DIR`> are the
install directories of the RCMake package, the RSys and the Embree libraries,
respectively. The resulting project can be edited, built, tested and installed
as any CMake project (Refer to the [CMake
documentation](https://cmake.org/documentation) for further informations on
CMake).

Example on a GNU/Linux system:

    ~ $ git clone https://gitlab.com/meso-star/star-2d.git
    ~ $ mkdir star-3d/build && cd star-2d/build
    ~/star-2d/build $ cmake -G "Unix Makefiles" \
    > -DCMAKE_PREFIX_PATH="<RCMAKE_DIR>;<RSYS_DIR>;<EMBREE_DIR>" \
    > -DCMAKE_INSTALL_PREFIX=<STAR2D_INSTALL_DIR> \
    > ../cmake
    ~/star-2d/build $ make && make test
    ~/star-2d/build $ make install

with `<STAR2D_INSTALL_DIR>` the directory in which Star-2D is going to be
installed.

## Quick Start

Once installed, the Star-2D library and its associated headers are deployed,
providing the whole environment required to develop C/C++ applications with
Star-2D.  The `<STAR2D_INSTALL_DIR>/include/star/s2d.h` header defines the
Star-2D Application Programming Interface (API). Refer to this
[file](https://gitlab.com/meso-star/star-2d/blob/master/src/s2d.h) for the API
reference documentation.


A Star-2D [CMake
package](https://cmake.org/cmake/help/v3.5/manual/cmake-packages.7.html) is
also installed to facilitate the use of Star-2D in projects relying on the
CMake build system. For instance, the following `CMakeLists.txt` file uses the
Star-2D package to build an executable relying on the Star-2D library.

    project(my_project C)

    # Use the Star-2D CMake package
    find_package(Star2D REQUIRED)

    # Define the program to build
    add_executable(my_program main.c)

    # Link the program against Star-2D
    target_link_libraries(my_program Star2D)

This `CMakeLists.txt` is then used to generate the target project as, for
instance, on a GNU/Linux system:

    cmake -G "Unix Makefiles" -DCMAKE_PREFIX_PATH="<STAR2D_INSTALL_DIR>" <MY_PROJECT_DIR>

with `<STAR2D_INSTALL_DIR>` the install directory of Star-2D and
`<MY_PROJECT_DIR>` the location of the `CMakeLists.txt` file.

## Release notes

### Version 0.5

Update the API of the filtering function: add the range of the ray as input
argument. For closest point queries, this range is from 0 to query radius.

### Vesion 0.4.1

- Improve the accuracy of the closest point query.
- Fix the closest point query: the returned hit was wrong.
- Fix the barycentric coordinates of the intersection of the ray onto the
  segment: the coordinates could lie outside the segment.

### Version 0.4

- Add the `s2d_scene_view_closest_point` function. This function retrieves the
  closest point into the scene to a query position in a radius from ]0, INF].
  This closest point is returned as a regular hit which means that the function
  returns to the caller not only the distance to the query position, but also
  the geometric primitive onto which the point lies and its location onto this
  primitive. If no segment lies around the query position in a distance lesser
  than the one defined by the query radius, the returned hit is invalid which
  can be checked with the regular `S2D_HIT_NONE` macro. Finally, this function
  supports the filter functions so that the caller can choose to discard
  closest hits according to its own criteria.

### Version 0.3.1

- Fix the `s2d_segment_get_vertex_attrib` function: it could fail and return an
  error while it shouldn't.

### Version 0.3

- Migrate the ray-tracing backend from Embree2 to Embree3.

### Version 0.2

- Add the `s2d_segment_get_vertex_attrib` function that returns the vertex
  attributes of a given segment.

### Version 0.1.1

- Fix the `s2d_primitive_get_function`. The function failed when the submitted
  parametric coordinate was equal to 1.

### Version 0.1

- Implement the `s2d_scene_view` API; it replaces the
  `s2d_scene_<begin|end>_session` functions that were removed. A view registers
  the state of the scene from which it is created. It is used to retrieve the
  scene data through ray-tracing, sampling or indexing. Several views can be
  created on the same scene.
- Add the possibility to attach a same shape to several scenes.

## License

Copyright (C) 2016-2021 |Meso|Star> (<contact@meso-star.com>). Star-2D is free
software released under the CeCILL v2.1 license. You are welcome to
redistribute it under certain conditions; refer to the COPYING files for
details.

